from typing import Sequence
import numpy as np

from quple import TemplateCircuitBlock

class ExcitationPreservingBlock(TemplateCircuitBlock):
    def __init__(self, mode:str='iswap'):
        supported_modes = ['iswap', 'fsim']
        if mode not in supported_modes:
            raise ValueError('Unsupported mode {}, choose one of {}'.format(mode, supported_modes))
        self.mode = mode
    
    @staticmethod
    def RYY(circuit:'ParameterisedCircuit', theta, qubits:Sequence[int]):
        circuit.RX(np.pi/2, list(qubits))
        circuit.CX(tuple(qubits))
        circuit.RZ(theta, qubits[1])
        circuit.CX(tuple(qubits))
        circuit.RX(-np.pi/2, list(qubits))
        
    @staticmethod
    def RXX(circuit:'ParameterisedCircuit', theta, qubits:Sequence[int]):
        circuit.H(list(qubits))
        circuit.CX(tuple(qubits))
        circuit.RZ(theta, qubits[1])
        circuit.CX(tuple(qubits))
        circuit.H(list(qubits))
        

    def build(self, circuit:'ParameterisedCircuit', qubits:Sequence[int]):
        theta = circuit.new_param()
        ExcitationPreservingBlock.RXX(circuit, theta, qubits)
        ExcitationPreservingBlock.RYY(circuit, theta, qubits)
        if self.mode == 'fsim':
            phi = circuit.new_param()
            circuit.CZ(phi, qubits)    
    
    @property
    def num_params(self) -> int:
        return 2
    
    @property
    def num_block_qubits(self) -> int:
        return 2


from typing import List, Union, Optional, Callable, Sequence, Tuple

from quple import ParameterisedCircuit
from quple.circuit.templates.excitation_preserving_block import ExcitationPreservingBlock

class ExcitationPreserving(ParameterisedCircuit):
    '''Excitation preserving trial wavefunction

    '''    
    def __init__(self, n_qubit: int, copies: int=2, 
                 mode:str='iswap',
                 entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                 parameter_symbol:str='θ', name:str=None):
        
        supported_modes = ['iswap', 'fsim']
        if mode not in supported_modes:
            raise ValueError('Unsupported mode {}, choose one of {}'.format(mode, supported_modes))
        
        excitation_preserving_block = ExcitationPreservingBlock(mode)   
        super().__init__(n_qubit=n_qubit, copies=copies,
                         rotation_blocks='RZ',
                         entanglement_blocks=excitation_preserving_block,
                         entangle_strategy=entangle_strategy,
                         parameter_symbol=parameter_symbol,
                         final_rotation_layer=False, name=name)